"""Calculation of statistics that require a pos-tagger in the pipeline"""

from spacy.tokens import Doc, Span
from spacy.language import Language
from typing import Counter, Union


@Language.factory("pos_stats", default_config={"use_pos": True})
def create_pos_stats_component(nlp: Language, name: str, use_pos: bool) -> Language:
    """Allows PosStats to be added to a spaCy pipe using nlp.add_pipe("pos_stats")"""

    tagger = {"tagger", "attribute_ruler"}
    if not tagger.intersection(set(nlp.pipe_names)):
        raise ValueError(
            "The pipeline does not contain a component for POS tagging. Please load a spaCy model which includes a 'tagger' or an 'attribute ruler' component."
        )
    return POSStatistics(nlp, use_pos=use_pos)


class POSStatistics:
    """spaCy v.3.0 component that adds attributes for POS statistics to `Doc` and `Span` objects."""

    def __init__(self, nlp: Language, use_pos: bool):
        """Initialise components"""
        self.use_pos = use_pos

        if not Doc.has_extension("pos_proportions"):
            Doc.set_extension("pos_proportions", getter=self.pos_proportions)

        if not Span.has_extension("pos_proportions"):
            Span.set_extension("pos_proportions", getter=self.pos_proportions)

    def __call__(self, doc):
        """Run the pipeline component"""
        return doc

    def pos_proportions(self, text: Union[Doc, Span]) -> dict:
        """
        Returns:
            Dict containing {pos_prop_POSTAG: proportion of all tokens tagged with POSTAG. Does not create a key if no tokens in the document fit the POSTAG.
        """
        pos_counts = Counter()
        if self.use_pos:
            pos_counts.update([token.pos_ for token in text])
        else:
            pos_counts.update([token.tag_ for token in text])
        pos_proportions = {
            "pos_prop_" + tag: count / len(text) for tag, count in pos_counts.items()
        }

        return pos_proportions



import textdescriptives as td
import spacy
import pytest


@pytest.fixture(scope="function")
def nlp():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives")
    return nlp


def test_extract_df_single_doc(nlp):
    doc = nlp("This is just a cute little text. Actually, it's two sentences.")
    td.extract_df(doc)
    for metric in ["descriptive_stats", "readability", "dependency_distance", "quality"]:
        td.extract_df(doc, metrics=metric)


def test_extract_df_pipe(nlp):
    text = [
        "I wonder how well the function works on multiple documents",
        "Very exciting to see, don't you think?",
    ]
    docs = nlp.pipe(text)
    td.extract_df(docs)


def test_extract_df_subsetters(nlp):
    doc = nlp("This is just a cute little text. Actually, it's two sentences.")
    df = td.extract_df(doc, include_text=False)
    assert "token_length_mean" not in df[td.readability_cols].columns
    assert "token_length_mean" not in df[td.dependency_cols].columns
    assert "lix" not in df[td.descriptive_stats_cols].columns


def test_extract_df_error(nlp):
    doc = nlp("Very brief text")

    with pytest.raises(Exception) as e_info:
        td.extract_df("This is just a string")
    with pytest.raises(Exception) as e_info:
        td.extract_df(doc, metrics="not a metric")
    with pytest.raises(Exception) as e_info:
        td.extract_df(doc, metrics=True)


def test_extract_dict_single_doc(nlp):
    doc = nlp("This is just a cute little text. Actually, it's two sentences.")
    td.extract_dict(doc)
    for metric in ["descriptive_stats", "readability", "dependency_distance"]:
        td.extract_dict(doc, metrics=metric)


def test_extract_df_pipe(nlp):
    text = [
        "I wonder how well the function works on multiple documents",
        "Very exciting to see, don't you think?",
    ]
    docs = nlp.pipe(text)
    assert len(td.extract_dict(docs)["token_length_mean"]) == 2



import pytest

from textdescriptives.components import Readability
from .books import *

from spacy.lang.en import English

import numpy as np
import ftfy


@pytest.fixture(scope="function")
def nlp():
    nlp = English()
    nlp.add_pipe("readability")
    return nlp


def test_readability_integration(nlp):
    assert "readability" == nlp.pipe_names[-1]


def test_readability(nlp):
    doc = nlp("This is a short and simple sentence")
    assert doc._.readability


@pytest.mark.parametrize(
    "text,expected",
    [
        ("", np.nan),
        ("#", np.nan),
    ],
)
def test_readability_edge(text, expected, nlp):
    doc = nlp(text)
    if np.isnan(expected):
        for v in doc._.readability.values():
            assert np.isnan(v)
    else:
        for v in doc._.readability.values():
            assert v == expected


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 72.80),
        (secret_garden, 92.73),
        (flatland, 68.18),
        (grade_1, 116.08),
        (grade_2, 98.18),
        (grade_3, 101.51),
        (grade_4, 102.08),
        (grade_6, 97.96),
        (grade_8, 91.68),
        (grade_10, 76.67),
        (grade_12, 73.72),
        (grade_14, 65.25),
    ],
)
def test_flesch_reading_ease(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert pytest.approx(expected, rel=1e-2) == doc._.readability["flesch_reading_ease"]


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 9.48),
        (secret_garden, 4.04),
        (flatland, 12.05),
        (grade_1, -1.65),
        (grade_2, 1.85),
        (grade_3, 1.65),
        (grade_4, 1.20),
        (grade_6, 2.63),
        (grade_8, 4.99),
        (grade_10, 6.75),
        (grade_12, 6.42),
        (grade_14, 10.04),
    ],
)
def test_flesch_kincaid_grade(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert (
        pytest.approx(expected, rel=1e-2) == doc._.readability["flesch_kincaid_grade"]
    )


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 9.47),
        (secret_garden, 6.63),
        (flatland, 9.91),
    ],
)
def test_smog(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert pytest.approx(expected, rel=1e-2) == doc._.readability["smog"]


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 12.19),
        (secret_garden, 6.98),
        (flatland, 15.05),
    ],
)
def test_gunning_fog(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert pytest.approx(expected, rel=1e-2) == doc._.readability["gunning_fog"]


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 11.99),
        (secret_garden, 5.40),
        (flatland, 14.98),
    ],
)
def test_automated_readability_index(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert (
        pytest.approx(expected, rel=1e-2)
        == doc._.readability["automated_readability_index"]
    )


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 8.75),
        (secret_garden, 6.39),
        (flatland, 7.91),
    ],
)
def test_coleman_liau_index(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert pytest.approx(expected, rel=1e-2) == doc._.readability["coleman_liau_index"]


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 42.70),
        (secret_garden, 27.07),
        (flatland, 49.87),
    ],
)
def test_lix(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert pytest.approx(expected, rel=1e-2) == doc._.readability["lix"]


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 4.37),
        (secret_garden, 1.81),
        (flatland, 5.50),
    ],
)
def test_rix(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert pytest.approx(expected, rel=1e-2) == doc._.readability["rix"]



from .about import __title__, __version__  # noqa: F401
from .components import (  # noqa: F401
    DependencyDistance,
    DescriptiveStatistics,
    POSStatistics,
    Quality,
    Readability,
)
from .dataframe_extract import (  # noqa: F401
    dependency_cols,
    descriptive_stats_cols,
    extract_df,
    extract_dict,
    readability_cols,
)
from .load_components import TextDescriptives  # noqa: F401



"""These books and several tests are borrowed from https://github.com/mholtzscher/spacy_readability"""

oliver_twist = """Among other public buildings in a certain town, which for many reasons
it will be prudent to refrain from mentioning, and to which I will
assign no fictitious name, there is one anciently common to most towns,
great or small: to wit, a workhouse; and in this workhouse was born; on
a day and date which I need not trouble myself to repeat, inasmuch as
it can be of no possible consequence to the reader, in this stage of
the business at all events; the item of mortality whose name is
prefixed to the head of this chapter.

For a long time after it was ushered into this world of sorrow and
trouble, by the parish surgeon, it remained a matter of considerable
doubt whether the child would survive to bear any name at all; in which
case it is somewhat more than probable that these memoirs would never
have appeared; or, if they had, that being comprised within a couple of
pages, they would have possessed the inestimable merit of being the
most concise and faithful specimen of biography, extant in the
literature of any age or country.

Although I am not disposed to maintain that the being born in a
workhouse, is in itself the most fortunate and enviable circumstance
that can possibly befall a human being, I do mean to say that in this
particular instance, it was the best thing for Oliver Twist that could
by possibility have occurred.  The fact is, that there was considerable
difficulty in inducing Oliver to take upon himself the office of
respiration,--a troublesome practice, but one which custom has rendered
necessary to our easy existence; and for some time he lay gasping on a
little flock mattress, rather unequally poised between this world and
the next: the balance being decidedly in favour of the latter.  Now,
if, during this brief period, Oliver had been surrounded by careful
grandmothers, anxious aunts, experienced nurses, and doctors of
profound wisdom, he would most inevitably and indubitably have been
killed in no time.  There being nobody by, however, but a pauper old
woman, who was rendered rather misty by an unwonted allowance of beer;
and a parish surgeon who did such matters by contract; Oliver and
Nature fought out the point between them.  The result was, that, after
a few struggles, Oliver breathed, sneezed, and proceeded to advertise
to the inmates of the workhouse the fact of a new burden having been
imposed  upon the parish, by setting up as loud a cry as could
reasonably have been expected from a male infant who had not been
possessed of that very useful appendage, a voice, for a much longer
space of time than three minutes and a quarter.

As Oliver gave this first proof of the free and proper action of his
lungs, the patchwork coverlet which was carelessly flung over the iron
bedstead, rustled; the pale face of a young woman was raised feebly
from the pillow; and a faint voice imperfectly articulated the words,
'Let me see the child, and die.'

The surgeon had been sitting with his face turned towards the fire:
giving the palms of his hands a warm and a rub alternately.  As the
young woman spoke, he rose, and advancing to the bed's head, said, with
more kindness than might have been expected of him:

'Oh, you must not talk about dying yet.'

'Lor bless her dear heart, no!' interposed the nurse, hastily
depositing in her pocket a green glass bottle, the contents of which
she had been tasting in a corner with evident satisfaction.

'Lor bless her dear heart, when she has lived as long as I have, sir,
and had thirteen children of her own, and all on 'em dead except two,
and them in the wurkus with me, she'll know better than to take on in
that way, bless her dear heart!  Think what it is to be a mother,
there's a dear young lamb do.'

Apparently this consolatory perspective of a mother's prospects failed
in producing its due effect.  The patient shook her head, and stretched
out her hand towards the child.

The surgeon deposited it in her arms.  She imprinted her cold white
lips passionately on its forehead; passed her hands over her face;
gazed wildly round; shuddered; fell back--and died. They chafed her
breast, hands, and temples; but the blood had stopped forever.  They
talked of hope and comfort. They had been strangers too long.

'It's all over, Mrs. Thingummy!' said the surgeon at last.

'Ah, poor dear, so it is!' said the nurse, picking up the cork of the
green bottle, which had fallen out on the pillow, as she stooped to
take up the child.  'Poor dear!'

'You needn't mind sending up to me, if the child cries, nurse,' said
the surgeon, putting on his gloves with great deliberation. 'It's very
likely it _will_ be troublesome. Give it a little gruel if it is.'  He
put on his hat, and, pausing by the bed-side on his way to the door,
added, 'She was a good-looking girl, too; where did she come from?'

'She was brought here last night,' replied the old woman, 'by the
overseer's order.  She was found lying in the street.  She had walked
some distance, for her shoes were worn to pieces; but where she came
from, or where she was going to, nobody knows.'

The surgeon leaned over the body, and raised the left hand.  'The old
story,' he said, shaking his head: 'no wedding-ring, I see. Ah!
Good-night!'

The medical gentleman walked away to dinner; and the nurse, having once
more applied herself to the green bottle, sat down on a low chair
before the fire, and proceeded to dress the infant.

What an excellent example of the power of dress, young Oliver Twist
was!  Wrapped in the blanket which had hitherto formed his only
covering, he might have been the child of a nobleman or a beggar; it
would have been hard for the haughtiest stranger to have assigned him
his proper station in society.  But now that he was enveloped in the
old calico robes which had grown yellow in the same service, he was
badged and ticketed, and fell into his place at once--a parish
child--the orphan of a workhouse--the humble, half-starved drudge--to
be cuffed and buffeted through the world--despised by all, and pitied
by none.

Oliver cried lustily. If he could have known that he was an orphan,
left to the tender mercies of church-wardens and overseers, perhaps he
would have cried the louder."""

secret_garden = """
When Mary Lennox was sent to Misselthwaite Manor to live with her uncle
everybody said she was the most disagreeable-looking child ever seen.
It was true, too.  She had a little thin face and a little thin body,
thin light hair and a sour expression.  Her hair was yellow, and her
face was yellow because she had been born in India and had always been
ill in one way or another.  Her father had held a position under the
English Government and had always been busy and ill himself, and her
mother had been a great beauty who cared only to go to parties and
amuse herself with gay people.  She had not wanted a little girl at
all, and when Mary was born she handed her over to the care of an Ayah,
who was made to understand that if she wished to please the Mem Sahib
she must keep the child out of sight as much as possible.  So when she
was a sickly, fretful, ugly little baby she was kept out of the way,
and when she became a sickly, fretful, toddling thing she was kept out
of the way also.  She never remembered seeing familiarly anything but
the dark faces of her Ayah and the other native servants, and as they
always obeyed her and gave her her own way in everything, because the
Mem Sahib would be angry if she was disturbed by her crying, by the
time she was six years old she was as tyrannical and selfish a little
pig as ever lived.  The young English governess who came to teach her
to read and write disliked her so much that she gave up her place in
three months, and when other governesses came to try to fill it they
always went away in a shorter time than the first one.  So if Mary had
not chosen to really want to know how to read books she would never
have learned her letters at all.

One frightfully hot morning, when she was about nine years old, she
awakened feeling very cross, and she became crosser still when she saw
that the servant who stood by her bedside was not her Ayah.

"Why did you come?" she said to the strange woman.  "I will not let you
stay.  Send my Ayah to me."

The woman looked frightened, but she only stammered that the Ayah could
not come and when Mary threw herself into a passion and beat and kicked
her, she looked only more frightened and repeated that it was not
possible for the Ayah to come to Missie Sahib.

There was something mysterious in the air that morning.  Nothing was
done in its regular order and several of the native servants seemed
missing, while those whom Mary saw slunk or hurried about with ashy and
scared faces.  But no one would tell her anything and her Ayah did not
come.  She was actually left alone as the morning went on, and at last
she wandered out into the garden and began to play by herself under a
tree near the veranda.  She pretended that she was making a flower-bed,
and she stuck big scarlet hibiscus blossoms into little heaps of earth,
all the time growing more and more angry and muttering to herself the
things she would say and the names she would call Saidie when she
returned.

"Pig! Pig! Daughter of Pigs!" she said, because to call a native a pig
is the worst insult of all.

She was grinding her teeth and saying this over and over again when she
heard her mother come out on the veranda with some one.  She was with a
fair young man and they stood talking together in low strange voices.
Mary knew the fair young man who looked like a boy.  She had heard that
he was a very young officer who had just come from England.  The child
stared at him, but she stared most at her mother.  She always did this
when she had a chance to see her, because the Mem Sahib--Mary used to
call her that oftener than anything else--was such a tall, slim, pretty
person and wore such lovely clothes.  Her hair was like curly silk and
she had a delicate little nose which seemed to be disdaining things,
and she had large laughing eyes.  All her clothes were thin and
floating, and Mary said they were "full of lace." They looked fuller of
lace than ever this morning, but her eyes were not laughing at all.
They were large and scared and lifted imploringly to the fair boy
officer's face.

"Is it so very bad? Oh, is it?" Mary heard her say.

"Awfully," the young man answered in a trembling voice.  "Awfully, Mrs.
Lennox.  You ought to have gone to the hills two weeks ago."

The Mem Sahib wrung her hands.

"Oh, I know I ought!" she cried.  "I only stayed to go to that silly
dinner party.  What a fool I was!"

At that very moment such a loud sound of wailing broke out from the
servants' quarters that she clutched the young man's arm, and Mary
stood shivering from head to foot.  The wailing grew wilder and wilder.
"What is it? What is it?" Mrs. Lennox gasped.

"Some one has died," answered the boy officer.  "You did not say it had
broken out among your servants."

"I did not know!" the Mem Sahib cried.  "Come with me!  Come with me!"
and she turned and ran into the house.

After that, appalling things happened, and the mysteriousness of the
morning was explained to Mary.  The cholera had broken out in its most
fatal form and people were dying like flies.  The Ayah had been taken
ill in the night, and it was because she had just died that the
servants had wailed in the huts.  Before the next day three other
servants were dead and others had run away in terror.  There was panic
on every side, and dying people in all the bungalows.

During the confusion and bewilderment of the second day Mary hid
herself in the nursery and was forgotten by everyone.  Nobody thought
of her, nobody wanted her, and strange things happened of which she
knew nothing.  Mary alternately cried and slept through the hours.  She
only knew that people were ill and that she heard mysterious and
frightening sounds.  Once she crept into the dining-room and found it
empty, though a partly finished meal was on the table and chairs and
plates looked as if they had been hastily pushed back when the diners
rose suddenly for some reason.  The child ate some fruit and biscuits,
and being thirsty she drank a glass of wine which stood nearly filled.
It was sweet, and she did not know how strong it was.  Very soon it
made her intensely drowsy, and she went back to her nursery and shut
herself in again, frightened by cries she heard in the huts and by the
hurrying sound of feet.  The wine made her so sleepy that she could
scarcely keep her eyes open and she lay down on her bed and knew
nothing more for a long time.

Many things happened during the hours in which she slept so heavily,
but she was not disturbed by the wails and the sound of things being
carried in and out of the bungalow.

When she awakened she lay and stared at the wall.  The house was
perfectly still.  She had never known it to be so silent before.  She
heard neither voices nor footsteps, and wondered if everybody had got
well of the cholera and all the trouble was over.  She wondered also
who would take care of her now her Ayah was dead.  There would be a new
Ayah, and perhaps she would know some new stories.  Mary had been
rather tired of the old ones.  She did not cry because her nurse had
died.  She was not an affectionate child and had never cared much for
any one.  The noise and hurrying about and wailing over the cholera had
frightened her, and she had been angry because no one seemed to
remember that she was alive.  Everyone was too panic-stricken to think
of a little girl no one was fond of.  When people had the cholera it
seemed that they remembered nothing but themselves.  But if everyone
had got well again, surely some one would remember and come to look for
her.

But no one came, and as she lay waiting the house seemed to grow more
and more silent.  She heard something rustling on the matting and when
she looked down she saw a little snake gliding along and watching her
with eyes like jewels.  She was not frightened, because he was a
harmless little thing who would not hurt her and he seemed in a hurry
to get out of the room.  He slipped under the door as she watched him.

"How queer and quiet it is," she said.  "It sounds as if there were no
one in the bungalow but me and the snake."

Almost the next minute she heard footsteps in the compound, and then on
the veranda.  They were men's footsteps, and the men entered the
bungalow and talked in low voices.  No one went to meet or speak to
them and they seemed to open doors and look into rooms.  "What
desolation!" she heard one voice say.  "That pretty, pretty woman!  I
suppose the child, too.  I heard there was a child, though no one ever
saw her."

Mary was standing in the middle of the nursery when they opened the
door a few minutes later.  She looked an ugly, cross little thing and
was frowning because she was beginning to be hungry and feel
disgracefully neglected.  The first man who came in was a large officer
she had once seen talking to her father.  He looked tired and troubled,
but when he saw her he was so startled that he almost jumped back.

"Barney!" he cried out.  "There is a child here! A child alone! In a
place like this! Mercy on us, who is she!"

"I am Mary Lennox," the little girl said, drawing herself up stiffly.
She thought the man was very rude to call her father's bungalow "A
place like this!" "I fell asleep when everyone had the cholera and I
have only just wakened up.  Why does nobody come?"

"It is the child no one ever saw!" exclaimed the man, turning to his
companions.  "She has actually been forgotten!"

"Why was I forgotten?" Mary said, stamping her foot.  "Why does nobody
come?"

The young man whose name was Barney looked at her very sadly.  Mary
even thought she saw him wink his eyes as if to wink tears away.

"Poor little kid!" he said.  "There is nobody left to come."

It was in that strange and sudden way that Mary found out that she had
neither father nor mother left; that they had died and been carried
away in the night, and that the few native servants who had not died
also had left the house as quickly as they could get out of it, none of
them even remembering that there was a Missie Sahib.  That was why the
place was so quiet.  It was true that there was no one in the bungalow
but herself and the little rustling snake."""

flatland = """
I call our world Flatland, not because we call it so, but to make its
nature clearer to you, my happy readers, who are privileged to live in
Space.

Imagine a vast sheet of paper on which straight Lines, Triangles,
Squares, Pentagons, Hexagons, and other figures, instead of remaining
fixed in their places, move freely about, on or in the surface, but
without the power of rising above or sinking below it, very much like
shadows--only hard and with luminous edges--and you will then have a
pretty correct notion of my country and countrymen.  Alas, a few years
ago, I should have said "my universe":  but now my mind has been opened
to higher views of things.

In such a country, you will perceive at once that it is impossible that
there should be anything of what you call a "solid" kind; but I dare
say you will suppose that we could at least distinguish by sight the
Triangles, Squares, and other figures, moving about as I have described
them.  On the contrary, we could see nothing of the kind, not at least
so as to distinguish one figure from another.  Nothing was visible, nor
could be visible, to us, except Straight Lines; and the necessity of
this I will speedily demonstrate.

Place a penny on the middle of one of your tables in Space; and leaning
over it, look down upon it.  It will appear a circle.

But now, drawing back to the edge of the table, gradually lower your
eye (thus bringing yourself more and more into the condition of the
inhabitants of Flatland), and you will find the penny becoming more and
more oval to your view, and at last when you have placed your eye
exactly on the edge of the table (so that you are, as it were, actually
a Flatlander) the penny will then have ceased to appear oval at all,
and will have become, so far as you can see, a straight line.

The same thing would happen if you were to treat in the same way a
Triangle, or Square, or any other figure cut out of pasteboard.  As
soon as you look at it with your eye on the edge on the table, you will
find that it ceases to appear to you a figure, and that it becomes in
appearance a straight line.  Take for example an equilateral
Triangle--who represents with us a Tradesman of the respectable class.
Fig. 1 represents the Tradesman as you would see him while you were
bending over him from above; figs. 2 and 3 represent the Tradesman, as
you would see him if your eye were close to the level, or all but on
the level of the table; and if your eye were quite on the level of the
table (and that is how we see him in Flatland) you would see nothing
but a straight line.

When I was in Spaceland I heard that your sailors have very similar
experiences while they traverse your seas and discern some distant
island or coast lying on the horizon.  The far-off land may have bays,
forelands, angles in and out to any number and extent; yet at a
distance you see none of these (unless indeed your sun shines bright
upon them revealing the projections and retirements by means of light
and shade), nothing but a grey unbroken line upon the water.

Well, that is just what we see when one of our triangular or other
acquaintances comes toward us in Flatland.  As there is neither sun
with us, nor any light of such a kind as to make shadows, we have none
of the helps to the sight that you have in Spaceland.  If our friend
comes closer to us we see his line becomes larger; if he leaves us it
becomes smaller:  but still he looks like a straight line; be he a
Triangle, Square, Pentagon, Hexagon, Circle, what you will--a straight
Line he looks and nothing else.

You may perhaps ask how under these disadvantageous circumstances we
are able to distinguish our friends from one another: but the answer to
this very natural question will be more fitly and easily given when I
come to describe the inhabitants of Flatland.  For the present let me
defer this subject, and say a word or two about the climate and houses
in our country."""

grade_1 = """A train! A train!
A train! A train!
Could you, would you,
On a train?
Not on a train! Not in a tree!
Not in a car! Sam! Let me be!
I would not, could not, in a box.
I could not, would not, with a fox.
I will not eat them with a mouse.
I will not eat them in a house.
I will not eat them here or there.
I will not eat them anywhere.
I do not like green eggs and ham.
I do not like them, Sam-I-am."""

grade_2 = """Jonathan pushed back the big iron pot and stood up.
There were no bears. But up the path came his father, carrying his gun. And with
him were Jonathan's Uncle James and his Uncle Samuel, his Uncle John and his
Uncle Peter. Jonathan had never in all his life been so glad to see the uncles.
"Jonathan!'" said his father, "what a fright you have given us! Where have you
been all this time?"
"Coming over Hemlock Mountain," said Jonathan in a small voice. And he ran
right into his father's arms."""

grade_3 = """For months I had been telling myself that I would never put the Magic Finger
upon anyone again—not after what happened to my teacher, old Mrs. Winter.
Poor old Mrs. Winter.
One day we were in class, and she was teaching us spelling. "Stand up," she said
to me, "and spell kat."
"That's an easy one," I said. "K-a-t."
"You are a stupid little girl!" Mrs. Winter said.
"I am not a stupid little girl!" I cried. "I am a very nice little girl!"
"Go and stand in the corner," Mrs. Winter said.
Then I got cross, and I saw red, and I put the Magic Finger on Mrs. Winter good
and strong, and almost at once...
Guess what?
Whiskers began growing out of her face! They were long black whiskers, just
like the little ones you see on a kat, only much bigger. And how fast they grew!
Before we had time to think, they were out to her ears!"""

grade_4 = """The wheelbarrow picked up speed, so quickly that it sort of kicked up like a
whipped horse. I thought the handle was going to rip right out of my fingers.
"Hang on," I said.
"If I can," said Soup.
We were running now, full speed, smack down Sutter's Hill and heading full tilt
toward the party. Ahead of us, the giant pumpkin bounced around inside the bin
of the barrow. I felt like we'd stolen the moon.
"We're out of control!" yelled Soup.
"Turn it. Do anything, anything!"
"Can't."
The front door of the Baptist Church grew bigger and bigger, rushing toward us
like a mad monster. My feet hardly touched the ground. I was too frightened to
hang on much longer, yet frightened even more to let loose. Soup was screaming
and so was I.
"Stop," wailed Soup.
From the street, there was one step up to the door of the Baptist Church. The
door was closed."""

grade_6 = """"Brothers. What do you expect of me—to stand idly by while you burn my son?
My son has brought death to none of us. The scratches he gave us are not on our
bodies but our pride. Brothers. How if my son is burnt do I go back and face her
who lives with me in my house? How do I look in the eyes of his sisters who
think the rainbow arches over him? Brothers. It is easier for me to fight you all
than go back and say that Cuyloga stood by and did nothing while his brothers
in anger put his son to the fire."
With the quickness of Long Tail, the panther, he took his knife and cut the boy's
thongs. Then he stood there waiting for the attack, but none came. The warriors 
were too astonished. They watched, sullen and yet fascinated by the drama. This
was the great Cuyloga at his bravest that they looked upon, and none knew what
he would do next"""

grade_8 = """All day Buck brooded by the pool or roamed restlessly about the camp. Death,
as a cessation of movement, as a passing out and away from the lives of the
living, he knew, and he knew John Thornton was dead. It left a great void in him
somewhat akin to hunger, but a void which ached and ached, and which food
could not fill. At times when he paused to contemplate the carcasses of the
Yeehats, he forgot the pain of it; and at such times he was aware of a great pride
in himself—a pride greater than any he had yet experienced. He had killed man,
the noblest game of all, and he had killed in the face of the law of club and fang.
He sniffed the bodies curiously. They had died so easily. I was harder to kill a
husky dog than them. They were no match at all, were it not for their arrows and
spears and clubs. Thenceforward he would be unafraid of them except when
they bore in their hands their arrows, spears, and clubs."""

grade_10 = """Looking upward, I surveyed the ceiling of my prison. It was some thirty or forty
feet overhead, and constructed much as the side walls. In one of its panels a very
singular figure riveted my whole attention. It was the painted figure of Time as
he is commonly represented, save that, in lieu of a scythe, he held what, at a
casual glance, I supposed to be the pictured image of a huge pendulum, such as
we see on antique clocks. There was something, however, in the appearance of
this machine which caused me to regard it more attentively. While I gazed
directly upward at it (for its position was immediately over my own) I fancied
that I saw it in motion. In an instant afterward the fancy was confirmed. Its
sweep was brief, and of course slow. I watched it for some minutes somewhat in
fear, but more in wonder. Wearied at length with observing its dull movement, I
turned my eyes upon the other objects in the cell.
A slight noise attracted my notice, and, looking to the floor, I saw several
enormous rats traversing it. They had issued from the wall which lay just within
view to my right."""

grade_12 = """For the rest he lived solitary, but not misanthropic, with his books and his
collection, classing and arranging specimens, corresponding with entomologists
in Europe, writing up a descriptive catalogue of his treasures. Such was the
history of the man whom I had come to consult upon Jim's case without any
definite hope. Simply to hear what he would have to say would have been a
relief. I was very anxious, but I respected the intense, almost passionate,
absorption with which he looked at a butterfly, as though on the bronze sheen of
these frail wings, in the white tracings, in the gorgeous markings, he could see
other things, an image of something as perishable and defying destruction as
these delicate and lifeless tissues displaying a splendour unmarked by death.
"Marvellous!" he repeated, looking up at me. "Look! The beauty—but that is
nothing—look at the accuracy, the harmony. And so fragile! And so strong! And
so exact! This is Nature—the balance of colossal forces. Every star is so—and
every blade of grass stands so—the mighty Kosmos in perfect equilibrium
produces—this. This wonder; this masterpiece of Nature—the great artist.\""""

grade_14 = """It would have been in consonance with the spirit of Captain Vere should
he on this occasion have concealed nothing from the condemned one;
should he indeed have frankly disclosed to him the part he himself had
played in bringing about the decision, at the same time revealing his
actuated motives. On Billy's side it is not improbable that such a
confession would have been received in much the same spirit that
prompted it. Not without a sort of joy indeed he might have appreciated
the brave opinion of him implied in his captain making such a confidant
of him. Nor as to the sentence itself could he have been insensible that it
was imparted to him as to one not afraid to die. Even more may have 
Graded Text Samples
Impact Information. Page 9
been. Captain Vere in the end may have developed the passion
sometimes latent under an exterior stoical or indifferent. He was old
enough to have been Billy's father. The austere devotee of military duty,
letting himself melt back into what remains primeval in our formalised
humanity, may in the end have caught Billy to his heart, even as
Abraham may have caught young Isaac on the brink of resolutely
offering him up in obedience to the exacting behest. """



"""Adds all components to a spaCy pipeline"""
from .components import (
    Readability,
    DependencyDistance,
    DescriptiveStatistics,
    POSStatistics,
    Quality
)

from spacy.language import Language
from spacy.tokens import Doc


@Language.factory("textdescriptives")
def create_textdescriptives_component(nlp: Language, name: str):
    for component in [
        "descriptive_stats",
        "readability",
        "dependency_distance",
        "pos_stats",
        "quality"
    ]:
        nlp.add_pipe(component, last=True)
    return TextDescriptives(nlp)


class TextDescriptives:
    """
    Utility spaCy v3.0 component to add all functionality from the TextDescriptives
    package to a `Doc` objects. See `DescriptiveStatistics`, `Readability`, and `DependencyDistance`
    for more details.

    USAGE:
        >>> import spacy
        >>> from textdescriptives import TextDescriptives
        >>> nlp = spacy.load("en_core_web_sm")
        >>> nlp.add_pipe("textdescriptives")
        >>> doc = nlp("This is a the first sentence. Here's a second, slightly longer one.")
        >>> doc._.dependency_distance
        >>> doc._.readability
        >>> doc[0:4]._.token_length
    """

    def __init__(self, nlp: Language):
        """Don't do anything, just return the Doc"""

    def __call__(self, doc: Doc):
        """Run the pipeline component"""
        return doc



"""
Utility functions for calculating various text descriptives
"""
from typing import Any, Callable, Union

from pyphen import Pyphen
from spacy.tokens import Doc, Span, Token


def filtered_tokens(doc: Union[Doc, Span]):
    """Return words in document or span.
    Filters punctuation and words that start with an apostrophe (contractions)"""
    filtered_tokens = [
        word for word in doc if not word.is_punct and "'" not in word.text
    ]
    return filtered_tokens


def n_sentences(doc: Union[Doc, Span]):
    """Return number of sentences in the document"""
    return len(list(doc.sents))


def n_tokens(doc: Union[Doc, Span]):
    """Return number of words in the document."""
    return len(doc._._filtered_tokens)  # pylint: disable=protected-access


def n_syllables(doc: Doc):
    """
    Return number of syllables per token
    """

    dic = Pyphen(lang=doc.lang_)

    def count_syl(token: Token):
        word_hyphenated = dic.inserted(token.lower_)
        return max(1, word_hyphenated.count("-") + 1)

    return [
        count_syl(token)
        for token in doc._._filtered_tokens  # pylint: disable=protected-access
    ]


def span_getter_to_token_getter(
    span_getter: Callable[[Span], Any]
) -> Callable[[Token], Any]:
    """Converts a span getter to a token getter.

    Args:
        span_getter (Callable[[Span], Any]):
            The span getter function.

    Returns:
        Callable[[Token], Any]: The token getter function.
    """

    def token_getter(token):
        return span_getter(token.doc[token.i : token.i + 1])

    return token_getter


def span_getter_to_doc_getter(
    span_getter: Callable[[Span], Any]
) -> Callable[[Doc], Any]:
    """Converts a span getter to a document getter.

    Args:
        span_getter (Callable[[Span], Any]):
            The span getter function.

    Returns:
        Callable[[Doc], Any]: The document getter function.
    """

    def doc_getter(doc):
        return span_getter(doc[:])

    return doc_getter



"""Component for calculating quality metrics."""
from collections import Counter, defaultdict
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from spacy.language import Language
from spacy.tokens import Doc, Span

from textdescriptives.components.utils import span_getter_to_doc_getter

DEFAULT_QUALITY_THRESHOLDS = {
    "n_stop_words": (2, None),
    "alpha_ratio": (0.8, None),
    "mean_word_length": (3, 10),
    "doc_length": (10, 100_000),
    "symbol_#_2_word_ratio": (None, 0.1),
    "proportion_ellipsis": (None, 0.3),
    "proportion_bullet_points": (None, 0.8),
    "duplicate_line_chr_fraction": (None, 0.2),
    "duplicate_paragraph_chr_fraction": (None, 0.2),
    "duplicate_5-gram_chr_fraction": (None, 0.15),
    "duplicate_6-gram_chr_fraction": (None, 0.14),
    "duplicate_7-gram_chr_fraction": (None, 0.13),
    "duplicate_8-gram_chr_fraction": (None, 0.12),
    "duplicate_9-gram_chr_fraction": (None, 0.11),
    "duplicate_10-gram_chr_fraction": (None, 0.1),
    "top_2-gram_chr_fraction": (None, 0.20),
    "top_3-gram_chr_fraction": (None, 0.18),
    "top_4-gram_chr_fraction": (None, 0.16),
    "contains_lorem ipsum": False,
}


def n_stop_words(span: Span) -> int:
    """Count the number of stop words in a document.

    Args:
        span (Span): spaCy span object

    Returns:
        int: number of stop words
    """
    return sum(t.is_stop for t in span)


def mean_word_length(span: Span) -> float:
    """Calculate the mean word length of a document.

    Args:
        span (Span): spaCy span object

    Returns:
        float: mean word length
    """
    tokens_lengths = [len(t) for t in span]
    if tokens_lengths:
        return float(np.mean(tokens_lengths))
    return 0.0


def alpha_ratio(span: Span) -> float:
    """The percentage of spacy tokens in this document which contain
    at leat one alphabetic character.

    Args:
        span (Span): spaCy span object

    Returns:
        float: alpha ratio
    """

    def contains_alpha(token):
        for char in token.text:
            if char.isalpha():
                return True
        return False

    token_contains_alpha = [contains_alpha(token) for token in span]
    if token_contains_alpha:
        return float(np.mean(token_contains_alpha))
    return 0.0


def proportion_bullet_points(  # pylint: disable=dangerous-default-value
    span: Span, bullet_point: set = {"-", "*"}
) -> float:
    """Calculate the proportion of lines which start with a bullet points in a span.

    Args:
        span (Span): spaCy span object
        bullet_point (set): set of bullet points

    Returns:
        float: proportion of bullet points
    """
    # check if has extension _lines
    if not hasattr(span._, "lines"):
        lines = span.text.split("\n")
    else:
        lines = span._.lines
    line_starts_with_bullet = [
        line.strip().startswith(tuple(bullet_point)) for line in lines
    ]
    if line_starts_with_bullet:
        return float(np.mean(line_starts_with_bullet))
    return 0.0


def proportion_ellipsis(  # pylint: disable=dangerous-default-value
    span: Span, ellipsis: set = {"…", "..."}
) -> float:
    """Calculate the proportion line which ends with an ellipsis in a span.

    Args:
        span (Span): spaCy span object
        ellipsis (set): set of ellipsis

    Returns:
        float: proportion of ellipsis
    """
    if not hasattr(span._, "lines"):
        lines = span.text.split("\n")
    else:
        lines = span._.lines

    line_ends_with_ellipsis = [line.strip().endswith(tuple(ellipsis)) for line in lines]
    if line_ends_with_ellipsis:
        return float(np.mean(line_ends_with_ellipsis))
    return 0.0


def get_ranges(arr: np.ndarray) -> List[Tuple[int, int]]:
    """Get true ranges from boolean array, i.e.

    Example:
        >>> get_ranges(np.array([0, 1, 1, 0, 0, 1, 1]))
        [(1, 3), (5, 7)]
    """
    ranges = []
    start = None
    for i, val in enumerate(arr):
        if (val and start) is None:
            start = i
        elif not val and start is not None:
            ranges.append((start, i))
            start = None
    if start is not None:
        ranges.append((start, len(arr)))
    return ranges


def duplicate_paragraph_chr_fraction(span: Span) -> float:
    """Calculate the character fraction of duplicate paragraphs.

    Args:
        span (Span): spaCy span object

    Returns:
        float: The fraction of duplicate characters.
    """
    chr_len = len(span.text)
    if chr_len == 0:
        return 0.0

    if not hasattr(span._, "paragraphs"):
        paragraphs = span.text.split("\n\n")
    else:
        paragraphs = span._.paragraphs
    paragraph_counter = Counter(paragraphs)

    duplicate_chr = 0
    for t, c in paragraph_counter.items():
        if c > 1:
            duplicate_chr += len(t) * (c - 1)
    frac = duplicate_chr / chr_len
    return frac


def duplicate_line_chr_fraction(span: Span) -> float:
    """Calculate the character fraction of duplicate lines.

    Args:
        span (Span): spaCy span object

    Returns:
        float: The fraction of duplicate characters.
    """
    chr_len = len(span.text)
    if chr_len == 0:
        return 0.0

    if not hasattr(span._, "lines"):
        lines = span.text.split("\n")
    else:
        lines = span._.lines
    line_counter = Counter(lines)

    duplicate_chr = 0
    for t, c in line_counter.items():
        if c > 1:
            duplicate_chr += len(t) * (c - 1)
    frac = duplicate_chr / chr_len
    return frac


def symbol_2_word_ratio(span: Span, symbol: str) -> float:
    """Calculate the ratio of symbols to words in a span.

    Args:
        span (Span): spaCy span object
        ratio (float): ratio of symbols to words
        symbol (str): symbol to count

    Returns:
        float: ratio of symbols to words
    """
    n_symbol = span.text.count(symbol)
    n_words = sum(not (t.is_space or t.is_punct) for t in span)
    if n_words:
        return n_symbol / n_words
    return 0.0


def span_ngrams(span: Span, ngram_range: Tuple[int, int]) -> Dict[str, Counter]:
    """Calculate the counts of n-grams in the specified range.

    Args:
        span (Span): spaCy span object
        ngram_range (Tuple[int, int]): The n-gram range.

    Returns:
        Dict[int, Dict[str, int, List[Span]]]: A dictionary that for each n in the ngram
            range contains the counts of the n-grams as well as the spans of the
            n-grams.
    """
    max_len = len(span)
    lower, upper = ngram_range
    shingles_count = {
        n: defaultdict(lambda: {"count": 0, "span": []})
        for n in range(lower, upper + 1)
    }
    for i, _ in enumerate(span):
        for ngram_size in range(lower, upper + 1):
            end = i + ngram_size
            if not end > max_len:
                ngram_span = span[i:end]
                ngram = ngram_span.text
                shingles_count[ngram_size][ngram]["count"] += 1
                shingles_count[ngram_size][ngram]["span"].append(ngram_span)
    return shingles_count


def duplicate_ngram_fraction(
    span: Span,
    ngram_range: Tuple[int, int],
) -> Dict[int, float]:
    """calculates the character fraction of duplicate n-gram over the overall text,
    taking care not to count overlapping n-grams twice. This does not include spaces
    between the n-grams.

    Args:
        span (Span): spaCy span object
        ngram_range (Tuple[int, int], optional): The n-gram range.

    Returns:
        Dict[int, float]: the fraction of duplicate characters for each
            n-gram size
    """
    max_len = len(span)
    chr_len = len(span.text)
    if chr_len == 0:
        return {n: 0.0 for n in range(ngram_range[0], ngram_range[1] + 1)}
    shingles_count = span_ngrams(span, ngram_range)
    duplicate_chr_fraction = {}
    for ngram_size, ngrams in shingles_count.items():
        # create a boolean array of the same length as the text
        # where True indicates that the token is a duplicate
        is_duplicate = np.zeros(max_len, dtype=bool)
        # set duplicate tokens to True
        for ngram, count in ngrams.items():
            if count["count"] > 1:
                for ngram_span in count["span"]:
                    is_duplicate[ngram_span.start : ngram_span.end] = True

        duplicate_chars = 0
        # get duplicate ranges from boolean array
        for start, end in get_ranges(is_duplicate):
            _span = span[start:end]
            duplicate_chars += _span.end_char - _span.start_char
        duplicate_chr_fraction[ngram_size] = duplicate_chars / chr_len
    return duplicate_chr_fraction


def top_ngram_chr_fraction(
    span: Span,
    ngram_range: Tuple[int, int],
    min_count: int = 0,
) -> float:
    """Calculated whether the character fraction of the top n-grams is below the
    given thresholds

    Args:
        span (Span): spaCy span object
        ngram_range (Tuple[int, int], optional): Range of n grams to examine.
        min_count (int): Minimum count of n-grams to before an n-gram is considered
            a top n-gram. Defaults to 0.

    Returns:
        float: The fraction of the top n-grams.
    """
    chr_len = len(span.text)
    if chr_len == 0:
        return {n: 0.0 for n in range(ngram_range[0], ngram_range[1] + 1)}

    ngram_counter = span_ngrams(span, ngram_range=ngram_range)
    top_ngram_chr_frac = {}
    for n in ngram_counter:
        # find the top n-gram
        ngram, count_span = max(ngram_counter[n].items(), key=lambda x: x[1]["count"])
        count = count_span["count"]
        if count >= min_count:
            # calculate the fraction of the top n-gram
            top_ngram_chr_frac[n] = (len(ngram) * count) / chr_len
        else:
            top_ngram_chr_frac[n] = 0.0

    return top_ngram_chr_frac


def contains_string(span: Span, string: str) -> bool:
    """Check if a span contains a string.

    Args:
        span (Span): spaCy span object
        string (str): string to check for

    Returns:
        bool: True if span contains string
    """
    return string in span.text


class Quality:
    """spaCy component for adding text quality metrics to the `Doc` and `Span` objects.
    Extracts metrics and returns them as a dictionary as the ._.quality attribute.
    """

    def __init__(  # pylint: disable=dangerous-default-value
        self,
        nlp: Language,
        name: str,
        symbols: List[str],
        contains: List[str],
        top_ngram_range: Tuple[int, int],
        top_ngram_min_count: int,
        duplicate_n_gram_fraction_range: Tuple[int, int],
        quality_thresholds: Optional[
            Dict[str, Union[bool, Tuple[Optional[float], Optional[float]]]]
        ] = None,
        force: bool = False,
    ):  # noqa: D107
        """Initialise components"""
        self.name = name
        self.force = force
        self.symbols = symbols
        self.contains = contains
        self.top_ngram_range = top_ngram_range
        self.top_ngram_min_count = top_ngram_min_count
        self.duplicate_n_gram_fraction_range = duplicate_n_gram_fraction_range
        if quality_thresholds is None:
            quality_thresholds = DEFAULT_QUALITY_THRESHOLDS
        self.quality_thresholds = quality_thresholds

        self.getters = {
            # heuristic quality filters
            "n_stop_words": n_stop_words,
            "alpha_ratio": alpha_ratio,
            "mean_word_length": mean_word_length,
            "doc_length": len,
            "proportion_ellipsis": proportion_ellipsis,
            "proportion_bullet_points": proportion_bullet_points,
            # text repetition
            "duplicate_line_chr_fraction": duplicate_line_chr_fraction,
            "duplicate_paragraph_chr_fraction": duplicate_paragraph_chr_fraction,
            "duplicate_ngram_chr_fraction": partial(
                duplicate_ngram_fraction, ngram_range=duplicate_n_gram_fraction_range
            ),
            "top_ngram_chr_fraction": partial(
                top_ngram_chr_fraction,
                ngram_range=top_ngram_range,
                min_count=top_ngram_min_count,
            ),
        }
        # add symbol to word ratio
        for symbol in symbols:
            self.getters[f"symbol_{symbol}_2_word_ratio"] = partial(
                symbol_2_word_ratio, symbol=symbol
            )
        # add contains
        for string in contains:
            self.getters[f"contains_{string}"] = partial(contains_string, string=string)

        self.extensions = {
            "passed_quality_check": self.passed_quality_thresholds,
            "quality": self.quality_getter,
        }

        self.set_extensions()

    def __call__(self, doc: Doc):
        """Run the pipeline component"""
        return doc

    def quality_getter(self, span: Span) -> Dict[str, Union[float, int, bool]]:
        """Apply quality functions to doc

        Args:
            span (Span): spaCy span object

        Returns:
            Dict[str, Union[float, int, bool]]: dictionary of quality metrics
        """
        quality = {}
        for name, getter in self.getters.items():
            if name == "top_ngram_chr_fraction":
                chr_frac = getter(span)
                for n_gram, frac in chr_frac.items():
                    quality[f"top_{n_gram}-gram_chr_fraction"] = frac
            elif name == "duplicate_ngram_chr_fraction":
                chr_frac = getter(span)
                for n_gram, frac in chr_frac.items():
                    quality[f"duplicate_{n_gram}-gram_chr_fraction"] = frac
            else:
                quality[name] = getter(span)
        return quality

    def passed_quality_thresholds(self, span: Span) -> bool:
        """
        Checks whether a span passed the quality thresholds
        """
        quality = span._.quality
        for name, threshold in self.quality_thresholds.items():
            if name not in quality:
                raise KeyError(f"Quality metric {name} not found in doc._.quality")
            if isinstance(threshold, bool):
                if quality[name] != threshold:
                    return False
            elif isinstance(threshold, tuple) and len(threshold) == 2:
                if threshold[0] is not None and quality[name] < threshold[0]:
                    return False
                if threshold[1] is not None and quality[name] > threshold[1]:
                    return False
            else:
                raise ValueError(
                    f"Quality threshold {name} is not a bool, or "
                    + f"Tuple of length 2, but {type(threshold)}."
                )

        return True

    def set_extensions(self):
        """Set required extensions."""

        for ext_name, span_getter in self.extensions.items():
            doc_getter = span_getter_to_doc_getter(span_getter)

            if not Span.has_extension(ext_name) or self.force is True:
                Span.set_extension(ext_name, getter=span_getter, force=True)
            if not Doc.has_extension(ext_name) or self.force is True:
                Doc.set_extension(ext_name, getter=doc_getter, force=True)


@Language.factory(
    "quality",
    default_config={
        "symbols": ["#"],
        "contains": ["lorem ipsum"],
        "top_ngram_range": [2, 4],
        "top_ngram_min_count": 3,
        "duplicate_n_gram_fraction_range": [5, 10],
        "force": True,
        "quality_thresholds": None,
    },
)
def create_quality_component(  # pylint: disable=dangerous-default-value
    nlp: Language,
    name: str,
    symbols: List[str],
    contains: List[str],
    top_ngram_range: Tuple[int, int],
    top_ngram_min_count: int,
    duplicate_n_gram_fraction_range: Tuple[int, int],
    quality_thresholds: Optional[
        Dict[str, Union[bool, Tuple[Optional[float], Optional[float]]]]
    ] = None,
    force: bool = True,
) -> Callable[[Doc], Doc]:
    """Allows Quality to be added to a spaCy pipe using nlp.add_pipe("quality").

    Set the following extensions:
    - {Span/Doc}._.quality
    - {Span/Doc}._.passed_quality_check

    It is also possible to optionally set the following extensions:
    - {Span/Doc}._.lines
    - {Span/Doc}._.paragraphs

    These are used to calculate some of the quality metrics. They can be overwritten if
    you e.g. wish lines to be split on "\\r\\n" instead of "\\n".

    A large part of the quality metrics were proposed by [1] and [2] for filtering
    out low quality text from large text corpora.

    References:
    [1] Rae, J. W., Borgeaud, S., Cai, T., Millican, K., Hoffmann, J., Song, F., ... &
        Irving, G. (2021). Scaling language models: Methods, analysis & insights from
        training gopher. arXiv preprint arXiv:2112.11446.
    [2] Raffel, C., Shazeer, N., Roberts, A., Lee, K., Narang, S., Matena, M., ... &
        Liu, P. J. (2020). Exploring the limits of transfer learning with a unified
        text-to-text transformer. J. Mach. Learn. Res., 21(140), 1-67.

    Args:
        nlp (Language): spaCy language object
        name (str): name of the component
        symbols (List[str]): list of symbols for which to calculate the
            proportion the ratio of symbols to words. Defaults to ["#"].
        contains (List[str]): list of strings for which to check whether the
            document contains them. Defaults to ["lorem ipsum"].
        top_ngram_range (Tuple[int]): range of n-grams to calculate the
            proportion of the top n-gram. Defaults to [2, 4].
        top_ngram_min_count (int): minimum number of times a n-gram must occur to
            be considered a top n-gram. Defaults to 3.
        duplicate_n_gram_fraction_range (Tuple[int]): range of n-grams to
            calculate the proportion of duplicate n-grams. Defaults to [5, 10].
        quality_thresholds (Dict[str, Union[bool, Tuple[Union[int, float, None],
            Union[int, float, None]]]]): A dictionary of quality thresholds indicated by
            either a range (Tuple), wherein the first value is the lower bound and the
            second value is the upper bound. Lower and upper bounds can be None, in
            which case they are not checked. Alternatively, a boolean can be provided,
            checking if the quality metric is boolean. For example, if you  don't want
            documents containing `lorem ipsum`, to pass the quality check, you can set
            `quality_thresholds={"contains_lorem_ipsum": False}`. Similar if you want to
            set a upper bound on the `duplicate_5-gram_chr_fraction`, you can set
            `quality_thresholds={"duplicate_5-gram_chr_fraction": (None, 0.15)}`.
            Default values are set in
            `textdescriptives.components..quality.DEFAULT_QUALITY_THRESHOLDS`.
        force (bool): whether to overwrite existing extensions. Defaults to True.


    Returns:
        Callable[[Doc], Doc]: the spaCy component

    Example:
        >>> import spacy
        >>> from spacy_quality import Quality
        >>> nlp = spacy.blank(("en_core_web_sm")
        >>> nlp.add_pipe("quality")
        >>> doc = nlp("This is a test")
        >>> # extract quality metrics
        >>> doc._.quality
        >>> # check whether the document passed the quality thresholds
        >>> doc._.passed_quality_check
    """
    return Quality(
        nlp,
        name=name,
        symbols=symbols,
        contains=contains,
        top_ngram_range=top_ngram_range,
        top_ngram_min_count=top_ngram_min_count,
        duplicate_n_gram_fraction_range=duplicate_n_gram_fraction_range,
        quality_thresholds=quality_thresholds,
        force=force,
    )



"""Calculation of descriptive statistics."""
from spacy.tokens import Doc, Span
from spacy.language import Language
from typing import Union
import numpy as np

from .utils import filtered_tokens, n_tokens, n_syllables, n_sentences


@Language.factory("descriptive_stats")
def create_descriptive_stats_component(nlp: Language, name: str):
    """Allows DescriptiveStatistics to be added to a spaCy pipe using nlp.add_pipe("descriptive_stats").
    If the pipe does not contain a parser or sentencizer, the sentencizer component is silently added."""
    sentencizers = set(["sentencizer", "parser"])
    if not sentencizers.intersection(set(nlp.pipe_names)):
        nlp.add_pipe("sentencizer")  # add a sentencizer if not one in pipe
    return DescriptiveStatistics(nlp)


class DescriptiveStatistics:
    """spaCy v.3.0 component that adds attributes with desriptive statistics to `Doc` and `Span` objects.
    The attributes relate to token and sentence length, number of syllables, and counts of tokens and sentences.
    """

    def __init__(self, nlp: Language):
        """Initialise components"""

        extensions = [
            "_n_sentences",
            "_n_tokens",
            "_n_syllables",
            "token_length",
            "sentence_length",
            "syllables",
            "counts",
        ]
        ext_funs = [
            n_sentences,
            n_tokens,
            n_syllables,
            self.token_length,
            self.sentence_length,
            self.syllables,
            self.counts,
        ]
        for ext, fun in zip(extensions, ext_funs):
            if ext not in ["_n_sentences", "sentence_length", "syllables"]:
                if not Span.has_extension(ext):
                    Span.set_extension(ext, getter=fun)
            if not Doc.has_extension(ext):
                Doc.set_extension(ext, getter=fun)

        if not Doc.has_extension("_filtered_tokens"):
            Doc.set_extension("_filtered_tokens", default=[])
        if not Span.has_extension("_filtered_tokens"):
            Span.set_extension("_filtered_tokens", getter=filtered_tokens)

    def __call__(self, doc):
        """Run the pipeline component"""
        doc._._filtered_tokens = filtered_tokens(doc)
        return doc

    def token_length(self, doc: Union[Doc, Span]):
        """Return dict with measures of token length"""
        token_lengths = [len(token) for token in doc._._filtered_tokens]
        return {
            "token_length_mean": np.mean(token_lengths),
            "token_length_median": np.median(token_lengths),
            "token_length_std": np.std(token_lengths),
        }

    def sentence_length(self, doc: Doc):
        """Return dict with measures of sentence length"""
        # get length of filtered tokens per sentence
        tokenized_sentences = [
            [
                token.text
                for token in sent
                if not token.is_punct and "'" not in token.text
            ]
            for sent in doc.sents
        ]
        len_sentences = [len(sentence) for sentence in tokenized_sentences]
        return {
            "sentence_length_mean": np.mean(len_sentences),
            "sentence_length_median": np.median(len_sentences),
            "sentence_length_std": np.std(len_sentences),
        }

    def syllables(self, doc: Doc):
        """Return dict with measures of syllables per token"""
        n_syllables = doc._._n_syllables
        return {
            "syllables_per_token_mean": np.mean(n_syllables),
            "syllables_per_token_median": np.median(n_syllables),
            "syllables_per_token_std": np.std(n_syllables),
        }

    def counts(self, doc: Union[Doc, Span], ignore_whitespace: bool = True):
        """Returns dict with keys: n_tokens, n_unique_tokens, proportion_unique_tokens,
        n_characters
        """
        n_tokens = doc._._n_tokens
        n_types = len(set([tok.lower_ for tok in doc._._filtered_tokens]))
        if ignore_whitespace:
            n_chars = len(doc.text.replace(" ", ""))
        else:
            n_chars = len(doc.text)

        if n_tokens == 0:
            prop_unique_tokens = np.nan
        else:
            prop_unique_tokens = n_types / n_tokens
        out = {
            "n_tokens": n_tokens,
            "n_unique_tokens": n_types,
            "proportion_unique_tokens": prop_unique_tokens,
            "n_characters": n_chars,
        }
        if isinstance(doc, Doc):
            out["n_sentences"] = doc._._n_sentences
        return out



from .dependency_distance import DependencyDistance  # noqa: F401
from .descriptive_stats import DescriptiveStatistics  # noqa: F401
from .pos_stats import POSStatistics  # noqa: F401
from .quality import Quality  # noqa: F401
from .readability import Readability  # noqa: F401



import spacy
import pytest
from textdescriptives.components import DependencyDistance
from .books import *
import numpy as np
import ftfy


@pytest.fixture(scope="function")
def nlp():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("dependency_distance")
    return nlp


def test_dependency_distance_integration(nlp):
    assert "dependency_distance" == nlp.pipe_names[-1]


def test_dependency_distance(nlp):
    doc = nlp("This is a short and simple sentence")
    assert doc._.dependency_distance
    assert doc[0:3]._.dependency_distance
    assert doc[0]._.dependency_distance


@pytest.mark.parametrize(
    "text,expected",
    [
        ("", np.nan),
        ("#", 0.0),
    ],
)
def test_dependency_distance_edge(text, expected, nlp):
    doc = nlp(text)
    for v in doc._.dependency_distance.values():
        if expected is np.nan:
            assert v is np.nan
        else:
            assert pytest.approx(expected, rel=1e-2) == v


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 3.19),
        (secret_garden, 2.29),
        (flatland, 3.33),
    ],
)
def test_mean_dep_distance(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert (
        pytest.approx(expected, rel=0.15)
        == doc._.dependency_distance["dependency_distance_mean"]
    )


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 1.45),
        (secret_garden, 0.88),
        (flatland, 1.20),
    ],
)
def test_std_dep_distance(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert (
        pytest.approx(expected, rel=0.15)
        == doc._.dependency_distance["dependency_distance_std"]
    )


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 0.395),
        (secret_garden, 0.41),
        (flatland, 0.44),
    ],
)
def test_mean_adj_dep(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert (
        pytest.approx(expected, rel=0.1)
        == doc._.dependency_distance["prop_adjacent_dependency_relation_mean"]
    )


@pytest.mark.parametrize(
    "text,expected",
    [
        (oliver_twist, 0.11),
        (secret_garden, 0.09),
        (flatland, 0.056),
    ],
)
def test_std_adj_dep(text, expected, nlp):
    text = ftfy.fix_text(text)
    text = " ".join(text.split())
    doc = nlp(text)
    assert (
        pytest.approx(expected, rel=0.1)
        == doc._.dependency_distance["prop_adjacent_dependency_relation_std"]
    )



"""Calculation of various readability metrics."""
from textdescriptives.components.utils import n_sentences
from spacy.tokens import Doc
from spacy.language import Language

from typing import Dict
import numpy as np

from .descriptive_stats import create_descriptive_stats_component


@Language.factory("readability")
def create_readability_component(nlp: Language, name: str):
    """Allows Readability to be added to a spaCy pipe using nlp.add_pipe("readability").
    Readability requires attributes from DescriptiveStatistics and adds it to the
    pipe if it not already loaded."""
    if "descriptive_stats" not in nlp.pipe_names:
        print(
            "'descriptive_stats' component is required for 'readability'. Adding to pipe."
        )
        nlp = nlp.add_pipe("descriptive_stats")
    return Readability(nlp)


class Readability:
    """spaCy v.3.0 component for adding readability metrics to `Doc` objects.
    Extracts metrics and returns them as a dictionary as the ._.readability attribute.
    """

    def __init__(self, nlp: Language):
        """Initialise components"""
        if not Doc.has_extension("readability"):
            Doc.set_extension("readability", getter=self.readability)

    def __call__(self, doc: Doc):
        """Run the pipeline component"""
        return doc

    def readability(self, doc: Doc) -> Dict[str, float]:
        """Apply readability functions and return a dict of the results."""
        hard_words = len([syllable for syllable in doc._._n_syllables if syllable >= 3])
        long_words = len([t for t in doc._._filtered_tokens if len(t) > 6])

        return {
            "flesch_reading_ease": self._flesch_reading_ease(doc),
            "flesch_kincaid_grade": self._flesch_kincaid_grade(doc),
            "smog": self._smog(doc, hard_words),
            "gunning_fog": self._gunning_fog(doc, hard_words),
            "automated_readability_index": self._automated_readability_index(doc),
            "coleman_liau_index": self._coleman_liau_index(doc),
            "lix": self._lix(doc, long_words),
            "rix": self._rix(doc, long_words),
        }

    def _flesch_reading_ease(self, doc: Doc):
        """
        206.835 - (1.015 X avg sent len) - (84.6 * avg_syl_per_word)

        Higher = easier to read

        Works best for English
        """
        avg_sentence_length = doc._.sentence_length["sentence_length_mean"]
        avg_syl_per_word = doc._.syllables["syllables_per_token_mean"]
        if avg_sentence_length == 0 or avg_syl_per_word == 0:
            return np.nan
        score = 206.835 - (1.015 * avg_sentence_length) - (84.6 * avg_syl_per_word)
        return score

    def _flesch_kincaid_grade(self, doc: Doc):
        """
        Score = grade required to read the text
        """
        avg_sentence_length = doc._.sentence_length["sentence_length_mean"]
        avg_syl_per_word = doc._.syllables["syllables_per_token_mean"]
        if avg_sentence_length == 0 or avg_syl_per_word == 0:
            return np.nan
        score = 0.39 * avg_sentence_length + 11.8 * avg_syl_per_word - 15.59
        return score

    def _smog(self, doc: Doc, hard_words: int):
        """
        grade level = 1.043( sqrt(30 * (hard words /n sentences)) + 3.1291

        Preferably need 30+ sentences. Will not work with less than 4
        """
        n_sentences = doc._._n_sentences
        if n_sentences >= 3:
            smog = (1.043 * (30 * (hard_words / n_sentences)) ** 0.5) + 3.1291
            return smog
        return np.nan

    def _gunning_fog(self, doc, hard_words: int):
        """
        Grade level = 0.4 * ((avg_sentence_length) + (percentage hard words))

        hard words = 3+ syllables
        """
        n_tokens = doc._._n_tokens
        if n_tokens == 0:
            return np.nan
        avg_sent_len = doc._.sentence_length["sentence_length_mean"]
        percent_hard_words = (hard_words / n_tokens) * 100
        return 0.4 * (avg_sent_len + percent_hard_words)

    def _automated_readability_index(self, doc: Doc):
        """
        Score = grade required to read the text
        """
        if len(doc) == 0:
            return np.nan
        score = (
            4.71 * doc._.token_length["token_length_mean"]
            + 0.5 * doc._.sentence_length["sentence_length_mean"]
            - 21.43
        )
        return score

    def _coleman_liau_index(self, doc: Doc):
        """
        score = 0.0588 * avg number of chars pr 100 words - 0.296 * avg num of sents pr 100 words -15.8

        Score = grade required to read the text
        """
        n_tokens = doc._._n_tokens
        if n_tokens == 0:
            return np.nan
        l = doc._.token_length["token_length_mean"] * 100
        s = (doc._._n_sentences / n_tokens) * 100
        return 0.0588 * l - 0.296 * s - 15.8

    def _lix(self, doc: Doc, long_words: int):
        """
        (n_words / n_sentences) + (n_words longer than 6 letters * 100) / n_words
        """
        n_tokens = doc._._n_tokens
        if n_tokens == 0:
            return np.nan
        percent_long_words = long_words / n_tokens * 100
        return doc._.sentence_length["sentence_length_mean"] + percent_long_words

    def _rix(self, doc: Doc, long_words: int):
        """n_long_words / n_sentences"""
        n_sentences = doc._._n_sentences
        if doc._._n_tokens == 0:
            return np.nan
        return long_words / n_sentences



"""Extract metrics as Pandas DataFrame"""
from spacy.tokens import Doc

from functools import reduce
from collections import defaultdict
from typing import Union, List
import types

import pandas as pd


class Extractor:
    def __init__(
        self,
        doc: Doc,
        metrics: Union[List[str], str] = "all",
        include_text: bool = True,
        as_dict=False,
    ):
        """Utility class to extract specified metrics to a Pandas DataFrame or dictionary

        Args:
            doc (Doc): a spaCy doc
            metrics (Union[list[str], str], optional): Which metrics to extract.
                One or more of ["descriptive_stats", "readability", "dependency_distance", "pos_stats", "all"].
                Defaults to "all".
            include_text (bool, optional): Whether to add a column containing the text. Defaults to True.
        """
        if not isinstance(doc, (Doc)):
            raise TypeError(f"doc should be a spaCy Doc object, not {type(doc)}.")

        valid_metrics = set(
            [
                "descriptive_stats",
                "readability",
                "dependency_distance",
                "pos_stats",
                "quality",
                "all",
            ]
        )
        if isinstance(metrics, str):
            metrics = [metrics]
        if not isinstance(metrics, list):
            raise TypeError(
                f"'metrics' should be string or list of strings, not {type(metrics)}"
            )
        if not set(metrics).issubset(valid_metrics):
            raise ValueError(
                f"'metrics' contained invalid metric.\nValid metrics are: ['all', 'descriptive_stats', 'readability', 'dependency_distance', 'pos_stats']"
            )

        self.include_text = include_text
        self.as_dict = as_dict

        if include_text:
            extraction = [self.__extract_text(doc)]
        else:
            extraction = []

        if "all" in metrics:
            for component in valid_metrics - set(["all"]):
                extraction.append(self.__unpack_extension(doc, component))
        else:
            for component in metrics:
                extraction.append(self.__unpack_extension(doc, component))

        if self.as_dict:
            self.out = reduce(lambda a, b: {**a, **b}, extraction)
        else:
            self.out = pd.concat(extraction, axis=1)

    def __get_descriptive_stats_dict(self, doc: Doc) -> pd.DataFrame:
        descriptive_stats = {
            **doc._.token_length,
            **doc._.sentence_length,
            **doc._.syllables,
            **doc._.counts,
        }
        return descriptive_stats

    def __unpack_extension(self, doc: Doc, extension: str) -> pd.DataFrame:
        """Unpacks the the values from the extension to a dict or dataframe

        Args:
            doc (Doc): Document to extract from
            extension (str): Extension to extract

        Returns:
            pd.DataFrame: DataFrame with extension values
        """
        # doc.get_extension returns a tuple of (default, method, getter, setter)
        # we only need the getter
        if extension == "descriptive_stats":
            values = self.__get_descriptive_stats_dict(doc)
        elif extension == "pos_stats":
            values = doc.get_extension('pos_proportions')[2](doc)
        else:
            values = doc.get_extension(extension)[2](doc)

        if self.as_dict:
            return values
        return pd.DataFrame.from_records([values])

    def __extract_text(self, doc: Doc) -> Union[pd.DataFrame, str]:
        if self.as_dict:
            return {"text": doc.text}
        return pd.DataFrame([doc.text], columns=["text"])


def extract_df(
    doc: Doc, metrics: Union[List[str], str] = "all", include_text: bool = True
) -> pd.DataFrame:
    """Extract calculated metrics from a spaCy Doc object or a generator of Docs from
    nlp.pipe to a Pandas DataFrame

    Args:
        doc (Doc): a spaCy doc or a generator of spaCy Docs
        metrics (Union[list[str], str], optional): Which metrics to extract.
                One or more of ["descriptive_stats", "readability", "dependency_distance", "pos_stats", "all"].
                Defaults to "all".
        include_text (bool, optional): Whether to add a column containing the text. Defaults to True.

    Returns:
        pd.DataFrame: DataFrame with a row for each doc and column for each metric.
    """
    if isinstance(doc, types.GeneratorType):
        rows = []
        for d in doc:
            metric_df = Extractor(d, metrics, include_text).out
            rows.append(metric_df)
        return pd.concat(rows, axis=0, ignore_index=True)
    return Extractor(doc, metrics, include_text).out


def extract_dict(
    doc: Doc, metrics: Union[List[str], str] = "all", include_text: bool = True
) -> dict:
    """Extract calculated metrics from a spaCy Doc object or a generator of Docs from
    nlp.pipe to a dictionary

    Args:
        doc (Doc): a spaCy doc or a generator of spaCy Docs
        metrics (Union[list[str], str], optional): Which metrics to extract.
                One or more of ["descriptive_stats", "readability", "dependency_distance", "pos_stats", "all"].
                Defaults to "all".
        include_text (bool, optional): Whether to add an entry containing the text. Defaults to True.

    Returns:
        dict: Dictionary with a key for each metric.
    """
    if isinstance(doc, types.GeneratorType):
        dict_list = []
        for d in doc:
            metric_dict = Extractor(d, metrics, include_text, as_dict=True).out
            dict_list.append(metric_dict)
        # concatenate values from each dict in list
        out = defaultdict(list)
        for d in dict_list:
            for key, value in d.items():
                out[key].append(value)
        return dict(out)
    return Extractor(doc, metrics, include_text, as_dict=True).out


"""Helpers to subset an extracted dataframe"""

readability_cols = [
    "flesch_reading_ease",
    "flesch_kincaid_grade",
    "smog",
    "gunning_fog",
    "automated_readability_index",
    "coleman_liau_index",
    "lix",
    "rix",
]

dependency_cols = [
    "dependency_distance_mean",
    "dependency_distance_std",
    "prop_adjacent_dependency_relation_mean",
    "prop_adjacent_dependency_relation_std",
]

descriptive_stats_cols = [
    "token_length_mean",
    "token_length_median",
    "token_length_std",
    "sentence_length_mean",
    "sentence_length_median",
    "sentence_length_std",
    "syllables_per_token_mean",
    "syllables_per_token_median",
    "syllables_per_token_std",
    "n_tokens",
    "n_unique_tokens",
    "proportion_unique_tokens",
    "n_sentences",
    "n_characters",
]



"""
Tests for the quality module.
"""

from typing import List, Tuple

import pytest
import spacy

from textdescriptives.components.quality import (
    alpha_ratio,
    duplicate_ngram_fraction,
    mean_word_length,
    n_stop_words,
    proportion_bullet_points,
    proportion_ellipsis,
    symbol_2_word_ratio,
    top_ngram_chr_fraction,
)


@pytest.fixture
def nlp():
    """Load a blank English model."""
    return spacy.blank("en")


@pytest.mark.parametrize(
    "text, stop_words",
    [
        ("", 0),
        ("This is a test.", 3),
        ("This is a test. This is a test.", 6),
    ],
)
def test_n_stop_words(text: str, stop_words: int, nlp: spacy.Language):
    """Test the n_stop_words function."""
    doc = nlp(text)
    assert n_stop_words(doc) == stop_words


# test mean word length
@pytest.mark.parametrize(
    "text, mean_length",
    [
        ("", 0),
        ("This is a test.", 2.4),
        ("This is a test. This is a test.", 2.4),
    ],
)
def test_mean_word_length(text: str, mean_length: float, nlp: spacy.Language):
    """Test the mean_word_length function."""
    doc = nlp(text)
    assert mean_word_length(doc) == mean_length


# test alpha ratio
@pytest.mark.parametrize(
    "text, alpha",
    [
        ("", 0),
        ("This is a test.", 0.8),
        ("This,, is a test. 12355 is €%&/( a <<s.", 0.38),
        ("This is a test. This is a test. 123", 0.72),
    ],
)
def test_alpha_ratio(text: str, alpha: float, nlp: spacy.Language):
    """Test the alpha_ratio function."""
    doc = nlp(text)
    assert abs(alpha_ratio(doc) - alpha) < 0.01


# test proportion bullet points
@pytest.mark.parametrize(
    "text, bullet_points",
    [
        ("", 0),
        ("- This is a test.", 1),
        ("- This is a test. \nThis is a test.", 0.5),
        ("* This is a test.", 1),
    ],
)
def test_proportion_bullet_points(text: str, bullet_points: float, nlp: spacy.Language):
    """Test the proportion_bullet_points function."""
    doc = nlp(text)
    assert abs(proportion_bullet_points(doc) - bullet_points) < 0.01


# test proportion ellipsis
@pytest.mark.parametrize(
    "text, ellipsis",
    [
        ("", 0),
        ("This is a test...", 1),
        ("This is a test. \nThis is a test...", 0.5),
    ],
)
def test_proportion_ellipsis(text: str, ellipsis: float, nlp: spacy.Language):
    """Test the proportion_ellipsis function."""
    doc = nlp(text)
    assert proportion_ellipsis(doc) == ellipsis


# test symbol to word ratio
@pytest.mark.parametrize(
    "text, symbol_2_word",
    [
        ("", 0),
        ("This is a test.", 0),
        ("# test symbol to word ratio", 0.2),
        ("#### test symbol to word ratio", 0.8),
    ],
)
def test_symbol_2_word_ratio(text: str, symbol_2_word: float, nlp: spacy.Language):
    """Test the symbol_2_word_ratio function."""
    doc = nlp(text)
    assert abs(symbol_2_word_ratio(doc, symbol="#") - symbol_2_word) < 0.01


# test duplicate ngram fraction
@pytest.mark.parametrize(
    "text, duplicate_ngram, ngram_range",
    [
        ("", (0, 0), (2, 3)),
        ("This is a test.", (0, 0), (2, 3)),
        ("This is a test. This is a test.", (1,), (2, 2)),
        ("This is a test. This is a test.", (1, 1), (2, 3)),
        ("This is a test. This is a test. This is not a test.", (0.90, 0.90), (2, 3)),
        (
            "This is a test. This is maybe a test. This is not a test.",
            (0.789, 0.649),
            (2, 3),
        ),
        ("This is a test. This is a test. This is a test.", (1,), (3, 3)),
        ("This is a test. This is a test. This is a test.", (1, 1, 1, 1), (2, 5)),
        ("This is a test, where there is no duplicate ngram.", (0, 0), (2, 3)),
        ("This is a test. This is a test.", (0,), (8, 8)),
    ],
)
def test_duplicate_ngram_chr_fraction(
    text: str,
    duplicate_ngram: List[float],
    ngram_range: Tuple[int, int],
    nlp: spacy.Language,
):
    """Test the duplicate_ngram_fraction function."""
    doc = nlp(text)
    d = duplicate_ngram_fraction(doc, ngram_range=ngram_range)
    assert d, "duplicate_ngram_fraction should not be empty"
    for i, j in zip(d, duplicate_ngram):  # , strict=True): # for python >3.10
        assert abs(d[i] - j) < 0.01


# test top ngram chr fraction
@pytest.mark.parametrize(
    "text, top_ngram_chr_frac, ngram_range",
    [
        ("", (0, 0), (2, 3)),
        ("This is a test.", (0.466, 0.6), (2, 3)),
        ("This is a test. This is a monkey", (0.437, 0.562, 0.437), (2, 4)),
        (
            "This is a test. This is a monkey. This is a star.",
            (0.428, 0.551, 0.449),
            (2, 4),
        ),
    ],
)
def test_top_ngram_chr_fraction(
    text: str,
    top_ngram_chr_frac: List[float],
    ngram_range: Tuple[int, int],
    nlp: spacy.Language,
):
    """Test the top_ngram_chr_fraction function."""
    doc = nlp(text)
    top_ngram_fractions = top_ngram_chr_fraction(doc, ngram_range=ngram_range)
    for i, j in zip(top_ngram_fractions.values(), top_ngram_chr_frac):
        assert abs(i - j) < 0.01


def test_quality_component(nlp: spacy.Language):
    """Test the quality component."""
    nlp.add_pipe("quality", config={"force": True})
    doc = nlp("This is a test. This is a test. This is a test.")
    assert doc._.quality["n_stop_words"] == 9
    assert doc._.quality["mean_word_length"] == 2.4
    assert doc._.quality["alpha_ratio"] == 0.8
    assert doc._.quality["proportion_bullet_points"] == 0
    assert doc._.quality["proportion_ellipsis"] == 0
    assert doc._.quality["symbol_#_2_word_ratio"] == 0
    assert doc._.quality["duplicate_5-gram_chr_fraction"] == 1
    assert abs(doc._.quality["top_2-gram_chr_fraction"] - 0.44) < 0.01
    assert doc._.passed_quality_check is False


def test_quality_component_with_config(nlp: spacy.Language):
    """Test the quality component with config."""

    quality_thresholds = {
        "n_stop_words": (3, None),
        "alpha_ratio": (None, 0.8),
        "mean_word_length": (1, 10),
        "doc_length": (10, 100_000),
        "symbol_._2_word_ratio": (None, 0.3),
        "proportion_ellipsis": (None, 0.3),
        "proportion_bullet_points": (None, 0.8),
        "duplicate_line_chr_fraction": (None, 0.2),
        "duplicate_paragraph_chr_fraction": (None, 0.2),
        "top_2-gram_chr_fraction": (None, 0.6),
        "top_3-gram_chr_fraction": (None, 0.6),
        "contains_lorem ipsum": False,
    }
    d = nlp.add_pipe(
        "quality",
        config={
            "symbols": ["."],
            "quality_thresholds": quality_thresholds,
            "force": True,
        },
    )

    doc = nlp("This is a test. This is a test. This is a test.")
    assert doc._.quality["n_stop_words"] == 9
    assert doc._.quality["mean_word_length"] == 2.4
    assert doc._.quality["alpha_ratio"] == 0.8
    assert doc._.quality["proportion_bullet_points"] == 0
    assert doc._.quality["proportion_ellipsis"] == 0
    assert doc._.quality["symbol_._2_word_ratio"] == 0.25
    assert doc._.quality["duplicate_5-gram_chr_fraction"] == 1
    assert doc._.quality["duplicate_8-gram_chr_fraction"] == 1
    assert abs(doc._.quality["top_3-gram_chr_fraction"] - 0.57) < 0.01
    assert doc._.passed_quality_check is True


@pytest.mark.parametrize(
    "text, passed",
    [
        ("", False),
        (
            "This is a reasonable text, which has a very good sentence structure and "
            + "will therefore pass the quality check.",
            True,
        ),
        (
            "This is repitious text, This is repitious text, This is repitious text.",
            False,
        ),
        ("This test has many symobls #!@#$%^&*()_+.", False),
        ("- this is a text of \n - bullet points", False),
    ],
)
def test_passed_quality_check(text: str, passed: bool, nlp: spacy.Language):
    """Test the passed_quality_check attribute."""
    nlp.add_pipe("quality", config={"force": True})
    doc = nlp(text)
    assert doc._.passed_quality_check == passed



import spacy
from spacy.tokens import Doc
from spacy.lang.en import English
import pytest
from textdescriptives.components import POSStatistics


@pytest.fixture(scope="function")
def nlp():
    nlp = spacy.load("en_core_web_sm", disable=("ner", "textcat"))
    nlp.add_pipe("pos_stats")

    return nlp


@pytest.fixture(scope="function")
def doc(nlp):
    words = [
        "Here",
        "is",
        "the",
        "first",
        "sentence",
        ".",
        "It",
        "was",
        "pretty",
        "short",
        ".",
        "Let",
        "'s",
        "make",
        "another",
        "one",
        "that",
        "'s",
        "slightly",
        "longer",
        "and",
        "more",
        "complex",
        ".",
    ]
    pos = [
        "ADV",
        "AUX",
        "DET",
        "ADJ",
        "NOUN",
        "PUNCT",
        "PRON",
        "AUX",
        "ADV",
        "ADJ",
        "PUNCT",
        "VERB",
        "PRON",
        "VERB",
        "DET",
        "NOUN",
        "PRON",
        "AUX",
        "ADV",
        "ADJ",
        "CCONJ",
        "ADV",
        "ADJ",
        "PUNCT",
    ]
    doc = Doc(
        nlp.vocab,
        words=words,
        pos=pos,
    )
    return doc


def test_pos_integrations(nlp):
    assert "pos_stats" == nlp.pipe_names[-1]


def test_pos_proportions_doc(doc):
    assert doc._.pos_proportions == pytest.approx(
        {
            "pos_prop_ADV": 0.1666,
            "pos_prop_AUX": 0.125,
            "pos_prop_DET": 0.083,
            "pos_prop_ADJ": 0.1666,
            "pos_prop_NOUN": 0.0833,
            "pos_prop_PUNCT": 0.125,
            "pos_prop_PRON": 0.125,
            "pos_prop_VERB": 0.083,
            "pos_prop_CCONJ": 0.0416,
        },
        rel=0.05,
    )


def test_pos_proportions_span(doc):
    span = doc[:]

    assert span._.pos_proportions == pytest.approx(
        {
            "pos_prop_ADV": 0.1666,
            "pos_prop_AUX": 0.125,
            "pos_prop_DET": 0.083,
            "pos_prop_ADJ": 0.1666,
            "pos_prop_NOUN": 0.0833,
            "pos_prop_PUNCT": 0.125,
            "pos_prop_PRON": 0.125,
            "pos_prop_VERB": 0.083,
            "pos_prop_CCONJ": 0.0416,
        },
        rel=0.01,
    )



import spacy
import pytest
from textdescriptives import TextDescriptives


@pytest.fixture(scope="function")
def nlp():
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("textdescriptives")
    return nlp


def test_integration(nlp):
    assert nlp.pipe_names[-1] == "textdescriptives"
    for component in [
        "descriptive_stats",
        "readability",
        "dependency_distance",
        "textdescriptives",
        "quality"
    ]:
        assert component in nlp.pipe_names


def test_simple(nlp):
    doc = nlp("This is a short and simple text")
    assert doc



from spacy.lang.en import English
import pytest
from textdescriptives.components import DescriptiveStatistics


@pytest.fixture(scope="function")
def nlp():
    nlp = English()
    nlp.add_pipe("sentencizer")
    nlp.add_pipe("descriptive_stats")
    return nlp


def test_descriptive_stats_integration(nlp):
    assert "descriptive_stats" == nlp.pipe_names[-1]


def test_descriptive_stats(nlp):
    doc = nlp("This is a short and simple sentence")
    assert doc._.token_length
    assert doc._.sentence_length
    assert doc._.syllables
    assert doc._.counts
    assert doc[0:3]._.token_length
    assert doc[0:3]._.counts


def test_token_length(nlp):
    doc = nlp("Gift cats your prey")
    assert doc._.token_length["token_length_mean"] == 4.0
    assert doc._.token_length["token_length_median"] == 4.0
    assert doc._.token_length["token_length_std"] == 0.0
    assert doc[0:2]._.token_length["token_length_mean"] == 4.0
    assert doc[0:2]._.token_length["token_length_median"] == 4.0
    assert doc[0:2]._.token_length["token_length_std"] == 0.0


def test_sentence_length(nlp):
    doc = nlp(
        "Here is the first sentence. It was pretty short. Let's make another one that's slightly longer and more complex."
    )
    assert (
        pytest.approx(6.33, rel=1e-2) == doc._.sentence_length["sentence_length_mean"]
    )
    assert (
        pytest.approx(5.0, rel=1e-3) == doc._.sentence_length["sentence_length_median"]
    )
    assert pytest.approx(2.62, rel=1e-2) == doc._.sentence_length["sentence_length_std"]


def test_syllables_simple(nlp):
    doc = nlp("These words are easy")
    assert doc._.syllables["syllables_per_token_mean"] == 1.0
    assert doc._.syllables["syllables_per_token_median"] == 1.0
    assert doc._.syllables["syllables_per_token_std"] == 0.0


def test_syllables_complex(nlp):
    doc = nlp("This sentence has complicated words in it")
    assert pytest.approx(1.43, rel=1e-2) == doc._.syllables["syllables_per_token_mean"]
    assert pytest.approx(1.0, rel=1e-3) == doc._.syllables["syllables_per_token_median"]
    assert pytest.approx(0.73, rel=1e-2) == doc._.syllables["syllables_per_token_std"]


def test_counts(nlp):
    doc = nlp(
        "Here is the first sentence. It was pretty short. Let's make another one that's slightly longer and more complex."
    )
    assert doc._.counts["n_tokens"] == 19
    assert doc._.counts["n_unique_tokens"] == 19
    assert doc._.counts["proportion_unique_tokens"] == 1.0
    assert doc._.counts["n_characters"] == 94
    assert doc._.counts["n_sentences"] == 3
    assert doc[0:6]._.counts["n_tokens"] == 5
    assert doc[0:6]._.counts["n_unique_tokens"] == 5
    assert doc[0:6]._.counts["proportion_unique_tokens"] == 1.0
    assert doc[0:6]._.counts["n_characters"] == 23


@pytest.mark.parametrize("text", ["", "#"])
def test_descriptive_edge(text, nlp):
    doc = nlp(text)
    assert doc._.token_length
    assert doc._.sentence_length
    assert doc._.syllables
    assert doc._.counts



"""Calculation of statistics related to dependency distance"""
from spacy.tokens import Doc, Token, Span
from spacy.language import Language

import numpy as np


@Language.factory("dependency_distance")
def create_dependency_distance_component(nlp: Language, name: str):
    """Create spaCy language factory that allows DependencyDistance attributes to be added to a pipe using nlp.add_pipe("dependency_distance")"""
    return DependencyDistance(nlp)


class DependencyDistance:
    """spaCy v.3.0 component that adds attributes to `Doc`, `Span`, and `Token` objects relating to dependency distance.
    Dependency distance can be used as a measure of syntactic complexity, and measures the distance from a word to its head word.
    For `Doc` objects, dependency distance is calculated on the sentence level."""

    def __init__(self, nlp: Language):
        """Initialise components"""
        if not Token.has_extension("dependency_distance"):
            Token.set_extension("dependency_distance", getter=self.token_dependency)
        if not Span.has_extension("dependency_distance"):
            Span.set_extension("dependency_distance", getter=self.span_dependency)
        if not Doc.has_extension("dependency_distance"):
            Doc.set_extension("dependency_distance", getter=self.doc_dependency)

    def __call__(self, doc: Doc):
        """Run the pipeline component"""
        return doc

    def token_dependency(self, token: Token) -> dict:
        """Token-level dependency distance"""
        dep_dist = 0
        ajd_dep = False
        if token.dep_ != "ROOT":
            dep_dist = abs(token.head.i - token.i)
            if dep_dist == 1:
                ajd_dep = True
        return {"dependency_distance": dep_dist, "adjacent_dependency": ajd_dep}

    def span_dependency(self, span: Span) -> dict:
        """Span-level aggregated dependency distance"""
        dep_dists, adj_deps = zip(
            *[token._.dependency_distance.values() for token in span]
        )
        return {
            "dependency_distance_mean": np.mean(dep_dists),
            "prop_adjacent_dependency_relation": np.mean(adj_deps),
        }

    def doc_dependency(self, doc: Doc) -> dict:
        """Doc-level dependency distance aggregated on sentence level"""
        if len(doc) == 0:
            return {
                "dependency_distance_mean": np.nan,
                "dependency_distance_std": np.nan,
                "prop_adjacent_dependency_relation_mean": np.nan,
                "prop_adjacent_dependency_relation_std": np.nan,
            }
        dep_dists, adj_deps = zip(
            *[sent._.dependency_distance.values() for sent in doc.sents]
        )
        return {
            "dependency_distance_mean": np.mean(dep_dists),
            "dependency_distance_std": np.std(dep_dists),
            "prop_adjacent_dependency_relation_mean": np.mean(adj_deps),
            "prop_adjacent_dependency_relation_std": np.std(adj_deps),
        }
